from __future__ import absolute_import, division, unicode_literals

import logging

from liknorm import LikNormMachine
from numpy import sign
from numpy.linalg import LinAlgError
from numpy_sugar import epsilon
from numpy_sugar.linalg import economic_qs

from optimix import FunctionReduce

from ..ep import EP


class ExpFamGP(EP, FunctionReduce):
    r"""Expectation Propagation for Generalised Gaussian Processes.

    Args:
        y (array_like): outcome variable.
        lik_name (str): likelihood name.
        mean (:class:`optimix.Function`): mean function.
                                          (Refer to :doc:`mean`.)
        cov (:class:`optimix.Function`): covariance function.
                                         (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.ggp import ExpFamGP
        >>> from glimix_core.lik import BernoulliProdLik
        >>> from glimix_core.link import LogitLink
        >>> from glimix_core.random import GGPSampler
        >>>
        >>> random = RandomState(1)
        >>>
        >>> lik = BernoulliProdLik(LogitLink())
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GGPSampler(lik, mean, cov).sample(random)
        >>>
        >>> ggp = ExpFamGP(y, 'bernoulli', mean, cov)
        >>> print('Before: %.4f' % ggp.feed().value())
        Before: -65.8090
        >>> ggp.feed().maximize(progress=False)
        >>> print('After: %.2f' % ggp.feed().value())
        After: -65.39
    """

    def __init__(self, y, lik_name, mean, cov):
        super(ExpFamGP, self).__init__()
        FunctionReduce.__init__(self, [mean, cov], name='ExpFamGP')

        self._logger = logging.getLogger(__name__)

        self._y = y
        self._mean = mean
        self._cov = cov

        self._mean_value = None
        self._cov_value = None

        self._machine = LikNormMachine(lik_name, 500)

    def __del__(self):
        if hasattr(self, '_machine'):
            self._machine.finish()

    def _compute_moments(self):
        tau = self._cav['tau']
        eta = self._cav['eta']
        self._machine.moments(self._y, eta, tau, self._moments)

    # def value(self, mean, cov):
    def value_reduce(self, values):  # pylint: disable=R0201
        r"""Log of the marginal likelihood.

        Args:
            mean (array_like): realised mean.
            cov (array_like): realised covariance.

        Returns:
            float: log of the marginal likelihood.
        """
        mean = values['ExpFamGP[0]']
        cov = values['ExpFamGP[1]']
        QS = economic_qs(cov)
        try:
            self._initialize(mean, (QS[0][0], QS[1]))
            self._params_update()
            lml = self._lml()
        except (ValueError, LinAlgError) as e:
            self._logger.info(str(e))
            lml = -1 / epsilon.small
        return lml

    # def gradient(self, mean, cov, gmean, gcov):  # pylint: disable=W0221
    def gradient_reduce(self, values, gradients):
        r"""Gradient of the log of the marginal likelihood.

        Args:
            mean (array_like): realised mean.
            cov (array_like): realised cov.
            gmean (array_like): realised mean derivative.
            gcov (array_like): realised covariance derivative.

        Returns:
            list: derivatives.
        """

        mean = values['ExpFamGP[0]']
        cov = values['ExpFamGP[1]']
        gmean = gradients['ExpFamGP[0]']
        gcov = gradients['ExpFamGP[1]']

        try:
            QS = economic_qs(cov)
            self._initialize(mean, (QS[0][0], QS[1]))
            self._params_update()

            grad = dict()

            for n, g in iter(gmean.items()):
                grad['ExpFamGP[0].' + n] = self._lml_derivative_over_mean(g)

            for n, g in iter(gcov.items()):
                QS = economic_qs(g)
                grad['ExpFamGP[1].' + n] = self._lml_derivative_over_cov(
                    (QS[0][0], QS[1]))

            return grad
        except (ValueError, LinAlgError) as e:
            self._logger.info(str(e))
            v = self.variables().select(fixed=False)
            return {i: -sign(v.get(i).value) / epsilon.small for i in v}



from __future__ import unicode_literals

from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.link import ProbitLink
from glimix_core.lik import (BernoulliProdLik, BinomialProdLik,
                                 DeltaProdLik, PoissonProdLik)

def test_delta_prod_lik():
    random = RandomState(0)

    lik = DeltaProdLik(ProbitLink())

    assert lik.name == 'Delta'

    lik.outcome = [1, 0, 1]
    assert_allclose(lik.outcome, [1, 0, 1])

    assert lik.sample_size == 3
    assert_allclose(lik.mean([-1, 0, 0.5]), [-1, 0, 0.5])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [-10, 0, 0.5])

def test_bernoulli_prod_lik():
    random = RandomState(0)

    lik = BernoulliProdLik(ProbitLink())

    assert lik.name == 'Bernoulli'

    lik.outcome = [1, 0, 1]
    assert_allclose(lik.outcome, [1, 0, 1])

    assert lik.sample_size == 3
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 1])

def test_binomial_prod_lik():
    random = RandomState(0)

    lik = BinomialProdLik([6, 2, 3], ProbitLink())
    assert_allclose(lik.ntrials, [6, 2, 3])

    assert lik.name == 'Binomial'

    lik.nsuccesses = [4, 0, 1]
    assert_allclose(lik.nsuccesses, [4, 0, 1])

    assert lik.sample_size == 3
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 2])

def test_poisson_prod_lik():
    random = RandomState(0)

    lik = PoissonProdLik(ProbitLink())

    assert lik.name == 'Poisson'

    lik.noccurrences = [1, 4, 3]
    assert_allclose(lik.noccurrences, [1, 4, 3])

    assert lik.sample_size == 3
    assert_allclose(lik.mean([-1, 0, 0.5]), [0.15865525, 0.5, 0.69146246])
    assert_allclose(lik.sample([-10, 0, 0.5], random), [0, 1, 1])



"""
***********************
glimix_core package
***********************

Fast inference for Generalized Linear Mixed Models.

"""

from __future__ import absolute_import as _absolute_import

from pkg_resources import DistributionNotFound as _DistributionNotFound
from pkg_resources import get_distribution as _get_distribution

from . import cov, ggp, glmm, gp, lik, link, lmm, mean, random, util

try:
    __version__ = _get_distribution('glimix_core').version
except _DistributionNotFound:
    __version__ = 'unknown'


def test():
    import os
    p = __import__('glimix_core').__path__[0]
    src_path = os.path.abspath(p)
    old_path = os.getcwd()
    os.chdir(src_path)

    try:
        return_code = __import__('pytest').main(['-q', '--doctest-modules'])
    finally:
        os.chdir(old_path)

    if return_code == 0:
        print("Congratulations. All tests have passed!")

    return return_code


__all__ = [
    'test', 'ggp', 'gp', 'lmm', 'glmm', 'cov', 'lik', 'mean', 'link', 'random',
    'util'
]



from __future__ import division

from numpy import zeros, maximum

class Site(object): # pylint: disable=R0903
    r"""EP parameters."""
    def __init__(self, n):
        self.tau = zeros(n)
        self.eta = zeros(n)

    def update(self, mean, variance, cav):
        self.tau[:] = maximum(1.0 / variance - cav['tau'], 0)
        self.eta[:] = mean / variance - cav['eta']



from __future__ import division

from numpy import zeros
from numpy.testing import assert_allclose

from glimix_core.cov import FreeFormCov
from optimix.testing import Assertion


def test_freeform_optimix():
    item0 = 0
    item1 = 1

    cov = FreeFormCov(2)
    cov.L = [[1, 0], [2, 1]]

    a = Assertion(lambda: cov, item0, item1, 0.0, Lu=zeros(3))

    a.assert_layout()
    a.assert_gradient()

    assert_allclose(cov.Lu, [1, 2, 1])
    cov.Lu = [1, 1, 1]
    assert_allclose(cov.Lu, [1, 1, 1])



from __future__ import division

from numpy import asarray, exp, log, newaxis, atleast_1d

from optimix import Function, Scalar


class EyeCov(Function):
    r"""Identity covariance function.

    The mathematical representation is

    .. math::

        f(\mathrm x_0, \mathrm x_1) = s \delta[\mathrm x_0 = \mathrm x_1]

    where :math:`s` is the scale parameter and :math:`\delta` is the Kronecker
    delta.
    """

    def __init__(self):
        Function.__init__(self, logscale=Scalar(0.0))

    @property
    def scale(self):
        r"""Scale parameter."""
        return exp(self.variables().get('logscale').value)

    @scale.setter
    def scale(self, scale):
        self.variables().get('logscale').value = log(scale)

    def value(self, x0, x1):
        r"""Covariance function evaluated at `(x0, x1)`.

        Args:
            x0 (array_like): left-hand side sample or samples.
            x1 (array_like): right-hand side sample or samples.

        Returns:
            :math:`s \delta[\mathrm x_0 = \mathrm x_1]`.
        """
        x0 = asarray(x0)
        x1 = asarray(x1)
        x0_ = atleast_1d(x0).ravel()[:, newaxis]
        x1_ = atleast_1d(x1).ravel()[newaxis, :]
        v = self.scale * (x0_ == x1_)
        return v.reshape(x0.shape + x1.shape)

    def gradient(self, x0, x1):
        r"""Derivative of the covariance function evaluated at `(x0, x1)`.

        Derivative of the covariance function over :math:`\log(s)`.

        Args:
            x0 (array_like): left-hand side sample or samples.
            x1 (array_like): right-hand side sample or samples.

        Returns:
            :math:`s \delta[\mathrm x_0 = \mathrm x_1]`.
        """
        return dict(logscale=self.value(x0, x1))



from numpy import array

from glimix_core.cov import LinearCov
from optimix.testing import Assertion


def test_linearcov_optimix():
    item0 = array([-1.5, 1.0])
    item1 = array([+1.24, 1.0])
    a = Assertion(LinearCov, item0, item1, 0.0, logscale=0.0)
    a.assert_layout()
    a.assert_gradient()



# pylint: disable=E1101
from __future__ import division

from numpy import dot, ones, stack, tril_indices_from, zeros, zeros_like

from optimix import Function, Vector


class FreeFormCov(Function):
    """
    General semi-definite positive matrix with no contraints.
    A free-form covariance matrix of dimension d has 1/2 * d * (d + 1) params
    """

    def __init__(self, size):
        """
        Args:
            dim:        dimension of the free-form covariance
            jitter:     extent of diagonal offset which is added for numerical stability
                        (default value: 1e-4)
        """
        tsize = ((size + 1) * size) / 2
        tsize = int(tsize)
        self._L = zeros((size, size))
        self._tril = tril_indices_from(self._L)
        self._L[self._tril] = 1
        Function.__init__(self, Lu=Vector(ones(tsize)))

    @property
    def L(self):
        self._L[self._tril] = self.variables().get('Lu').value
        return self._L

    @L.setter
    def L(self, value):
        self._L[:] = value
        self.variables().get('Lu').value = self._L[self._tril]

    @property
    def Lu(self):
        return self.variables().get('Lu').value

    @Lu.setter
    def Lu(self, value):
        self.variables().get('Lu').value = value

    def value(self, x0, x1):
        return dot(self.L, self.L.T)[x0, ...][..., x1]

    def gradient(self, x0, x1):
        L = self.L
        Lo = zeros_like(L)
        grad = []
        for ii in range(len(self._tril[0])):
            row = self._tril[0][ii]
            col = self._tril[1][ii]
            Lo[row, col] = 1
            grad.append(dot(Lo, L.T) + dot(L, Lo.T))
            Lo[row, col] = 0

        grad = [g[x0, ...][..., x1] for g in grad]
        return dict(Lu=stack(grad, axis=-1))



from __future__ import division

from numpy import arange, sqrt
from numpy.random import RandomState
from numpy.testing import assert_allclose

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.ggp import ExpFamGP
from glimix_core.lik import BernoulliProdLik
from glimix_core.link import LogitLink
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler
from optimix import check_grad


def _get_data():
    random = RandomState(458)
    N = 100
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(N), purpose='sample')
    mean.set_data(arange(N), purpose='learn')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')
    cov_left.set_data((X, X), purpose='learn')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose='sample')
    cov_right.set_data((arange(N), arange(N)), purpose='learn')

    cov = SumCov([cov_left, cov_right])

    lik = BernoulliProdLik(LogitLink())

    y = GGPSampler(lik, mean, cov).sample(random)

    return dict(
        mean=mean,
        cov=cov,
        lik=lik,
        y=y,
        cov_left=cov_left,
        cov_right=cov_right)


def test_expfam_ep():
    data = _get_data()
    ep = ExpFamGP((data['y'], ), 'bernoulli', data['mean'], data['cov'])
    assert_allclose(ep.feed().value(), -60.84029280372346)


def test_expfam_ep_function():
    import logging
    logging.basicConfig(level=logging.DEBUG)
    data = _get_data()
    ep = ExpFamGP((data['y'], ), 'bernoulli', data['mean'], data['cov'])

    assert_allclose(check_grad(ep.feed()), 0, atol=1e-4)


def test_expfam_ep_optimize():
    data = _get_data()
    ep = ExpFamGP((data['y'], ), 'bernoulli', data['mean'], data['cov'])
    data['cov_left'].fix('logscale')
    ep.feed().maximize(progress=False)
    assert_allclose(data['cov_right'].scale, 4.165865119892221e-06, atol=1e-5)
    assert_allclose(data['mean'].offset, 1.0326586373049373)



from numpy.random import RandomState
from numpy.testing import assert_allclose
from numpy import ascontiguousarray, sqrt, ones, dot, zeros
from numpy_sugar.linalg import economic_qs, economic_qs_linear, sum2diag

from glimix_core.example import linear_eye_cov
from glimix_core.glmm import GLMM
from glimix_core.random import bernoulli_sample

def get_data(n):
    random = RandomState(0)

    X = ones((n, 1))
    G = random.randn(n, n + 5)
    K = dot(G, G.T)
    K /= K.diagonal().mean()
    sum2diag(K, 1e-3, out=K)
    QS = economic_qs(K)

    ntri = random.randint(1, 30, n)

    z = dot(G, random.randn(n + 5)) / sqrt(n + 5)

    nsuc = zeros(len(ntri), int)
    for i in range(len(ntri)):
        for j in range(ntri[i]):
            nsuc[i] += int(z[i] + 0.5 * random.randn() > 0)

    return (ntri, nsuc, QS, X)

class TimeSuite(object):
    def setup(self):
        self._data = {10: get_data(10), 100: get_data(100)}

    def time_qep_binomial_10_learn(self):
        (ntri, nsuc, QS, X) = self._data[10]

        glmm = GLMM((nsuc, ntri), 'binomial', X, QS)
        glmm.feed().maximize(progress=False)
        assert_allclose(glmm.feed().value(), -19.74077399053363)

    def time_qep_binomial_100_learn(self):
        (ntri, nsuc, QS, X) = self._data[100]

        glmm = GLMM((nsuc, ntri), 'binomial', X, QS)
        glmm.feed().maximize(progress=False)
        assert_allclose(glmm.feed().value(), -218.9503853656612)



import numpy as np
from numpy.testing import assert_allclose

from glimix_core.cov import LinearCov, SumCov
from optimix import check_grad


def test_value():
    random = np.random.RandomState(0)
    cov_left = LinearCov()
    cov_right = LinearCov()

    X0 = random.randn(4, 20)
    X1 = random.randn(4, 15)

    cov_left.set_data((X0, X0))
    cov_right.set_data((X1, X1))

    cov = SumCov([cov_left, cov_right])
    K = cov.feed().value()
    assert_allclose(K[0, 0], 37.95568923)
    assert_allclose(K[3, 1], 4.53034295)


def test_gradient():
    random = np.random.RandomState(0)
    cov_left = LinearCov()
    cov_right = LinearCov()

    X0 = random.randn(4, 20)
    X1 = random.randn(4, 15)

    cov_left.set_data((X0, X0))
    cov_right.set_data((X1, X1))

    cov = SumCov([cov_left, cov_right])

    assert_allclose(check_grad(cov.feed()), 0, atol=6)



"""
***********
Likelihoods
***********

- Likelihood products: :class:`.DeltaProdLik`, :class:`.BernoulliProdLik`,
  :class:`.BinomialProdLik`, and :class:`.PoissonProdLik`.

Likelihood products
^^^^^^^^^^^^^^^^^^^

.. autoclass:: DeltaProdLik
  :members:

.. autoclass:: BernoulliProdLik
  :members:

.. autoclass:: BinomialProdLik
  :members:

.. autoclass:: PoissonProdLik
  :members:

"""

from .prod import (BernoulliProdLik, BinomialProdLik, DeltaProdLik,
                   PoissonProdLik)

__all__ = [
    'DeltaProdLik', 'BernoulliProdLik', 'BinomialProdLik', 'PoissonProdLik'
]



from __future__ import division

import pytest
from numpy import arange, concatenate, inf, nan, newaxis, ones, sqrt, zeros
from numpy.random import RandomState
from numpy.testing import assert_allclose
from numpy_sugar.linalg import economic_qs_linear

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.lik import DeltaProdLik
from glimix_core.lmm import LMM
from glimix_core.lmm.core import LMMCore
from glimix_core.mean import OffsetMean
from glimix_core.random import GGPSampler


def test_fastlmm_fast_scan():  # pylint: disable=R0914
    random = RandomState(9458)
    N = 500
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(N), purpose='sample')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose='sample')

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    y = GGPSampler(lik, mean, cov).sample(random)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((N, 1)), QS)

    lmm.learn(progress=False)

    markers = random.randn(N, 2)

    lmm_ = lmm.copy()
    lmm_.X = concatenate([lmm.X, markers[:, 0][:, newaxis]], axis=1)
    lmm_.fix('delta')
    lmm_.learn(progress=False)
    lml0 = lmm_.lml()

    lmm_ = lmm.copy()
    lmm_.X = concatenate([lmm.X, markers[:, 1][:, newaxis]], axis=1)
    lmm_.fix('delta')
    lmm_.learn(progress=False)
    lml1 = lmm_.lml()

    fast_scanner = lmm.get_fast_scanner()

    lmls = fast_scanner.fast_scan(markers)[0]
    assert_allclose(lmls, [lml0, lml1])


def test_lmm_learn():
    random = RandomState(9458)
    N = 500
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(N), purpose='sample')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose='sample')

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    y = GGPSampler(lik, mean, cov).sample(random)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((N, 1)), QS)

    lmm.learn(progress=False)

    assert_allclose(lmm.beta[0], 0.8997652129631661, rtol=1e-5)
    assert_allclose(lmm.genetic_variance, 1.7303981309775553, rtol=1e-5)
    assert_allclose(lmm.environmental_variance, 1.2950028351268132, rtol=1e-5)

    lmm.beta = [-0.5]
    assert_allclose(lmm.beta[0], [-0.5])


def test_fastlmm_learn_fix():
    random = RandomState(9458)
    N = 500
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])
    offset = 1.0

    mean = OffsetMean()
    mean.offset = offset
    mean.set_data(arange(N), purpose='sample')

    cov_left = LinearCov()
    cov_left.scale = 1.5
    cov_left.set_data((X, X), purpose='sample')

    cov_right = EyeCov()
    cov_right.scale = 1.5
    cov_right.set_data((arange(N), arange(N)), purpose='sample')

    cov = SumCov([cov_left, cov_right])

    lik = DeltaProdLik()

    y = GGPSampler(lik, mean, cov).sample(random)

    QS = economic_qs_linear(X)

    lmm = LMM(y, ones((N, 1)), QS)

    lmm.fix('delta')
    lmm.fix('scale')

    lmm.scale = 1.0
    lmm.delta = 0.5

    lmm.learn(progress=False)

    assert_allclose(lmm.beta[0], 0.899765212963)
    assert_allclose(lmm.scale, 1.0)
    assert_allclose(lmm.delta, 0.5)
    assert_allclose(lmm.genetic_variance, 0.5)
    assert_allclose(lmm.environmental_variance, 0.5)
    assert_allclose(lmm.lml(), -681.381571238)

    lmm.unfix('scale')
    lmm.learn(progress=False)

    assert_allclose(lmm.beta[0], 0.899765212963)
    assert_allclose(lmm.genetic_variance, 1.4614562029852856)
    assert_allclose(lmm.environmental_variance, 1.4614562029852856)
    assert_allclose(lmm.lml(), -949.526700867)

    lmm.unfix('delta')
    lmm.learn(progress=False)

    assert_allclose(lmm.beta[0], 0.899765212963)
    assert_allclose(lmm.genetic_variance, 1.73039821903)
    assert_allclose(lmm.environmental_variance, 1.29500280131)
    assert_allclose(lmm.lml(), -948.798268063)


def test_lmm_unique_outcome():
    random = RandomState(9458)
    N = 50
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    QS = economic_qs_linear(X)

    lmm = LMM(zeros(N), ones((N, 1)), QS)

    lmm.learn(progress=False)

    assert_allclose(lmm.beta[0], 0, atol=1e-7)
    assert_allclose(lmm.genetic_variance, 0, atol=1e-7)
    assert_allclose(lmm.environmental_variance, 0, atol=1e-7)


def test_lmm_nonfinite_phenotype():
    random = RandomState(9458)
    N = 50
    QS = economic_qs_linear(random.randn(N, N + 1))
    y = zeros(N)

    y[0] = nan
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)

    y[0] = -inf
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)

    y[0] = +inf
    with pytest.raises(ValueError):
        LMM(y, ones((N, 1)), QS)


def test_lmmcore_interface():
    random = RandomState(9458)
    N = 50
    QS = economic_qs_linear(random.randn(N, N + 1))
    y = zeros(N)

    lmmc = LMMCore(y, ones((N, 1)), QS)
    with pytest.raises(NotImplementedError):
        print(lmmc.delta)

    with pytest.raises(NotImplementedError):
        lmmc.delta = 1



r"""
*******************
Linear Mixed Models
*******************

Introduction
^^^^^^^^^^^^

A LMM can be described as

.. math::

    \mathbf y = \mathrm X\boldsymbol\beta + \mathrm G\mathbf u
    + \boldsymbol\epsilon,

where :math:`\mathbf u \sim \mathcal N(\mathbf 0, \sigma_r^2\mathrm I)` is a
vector of random effects and :math:`\epsilon_i` are iid Normal random variables
with zero-mean and variance :math:`\sigma_e^2` each.
The outcome-vector is thus distributed according to

.. math::

    \mathbf y \sim \mathcal N(\mathrm X\boldsymbol\beta,
                              \sigma_r^2 \mathrm G \mathrm G^{\intercal}
                              + \sigma_e^2\mathrm I)

This module provides :class:`.LMM`, a FastLMM implementation to perform
inference.
:class:`.FastScanner` can be used to perform fast inference over multiple
covariates.

Public interface
^^^^^^^^^^^^^^^^
"""

from .lmm import LMM
from .scan import FastScanner

__all__ = ['LMM', 'FastScanner']



def create_cache():
    from cachetools import LRUCache

    return LRUCache(maxsize=1)

def cached(method):
    from operator import attrgetter
    from cachetools import cachedmethod

    cache_name = '_cache_%s' % method.__name__
    return cachedmethod(attrgetter(cache_name))(method)



from numpy import arange, sqrt
from numpy.random import RandomState

from glimix_core.cov import EyeCov, LinearCov, SumCov
from glimix_core.mean import OffsetMean


def offset_mean():
    N = 100
    mean = OffsetMean()
    mean.offset = 0.5
    mean.set_data(arange(N), purpose='sample')
    mean.set_data(arange(N), purpose='learn')

    return mean


def linear_eye_cov():
    random = RandomState(458)
    N = 100
    X = random.randn(N, N + 1)
    X -= X.mean(0)
    X /= X.std(0)
    X /= sqrt(X.shape[1])

    cov_left = LinearCov()
    cov_left.scale = 1.0
    cov_left.set_data((X, X), purpose='sample')
    cov_left.set_data((X, X), purpose='learn')

    cov_right = EyeCov()
    cov_right.scale = 1.0
    cov_right.set_data((arange(N), arange(N)), purpose='sample')
    cov_right.set_data((arange(N), arange(N)), purpose='learn')

    return SumCov([cov_left, cov_right])



r"""
*******************************
Generalized Linear Mixed Models
*******************************

Introduction
^^^^^^^^^^^^

A GLMM can be described in two parts.
The first part consists in the latent variable

.. math::

    \mathbf z = \mathrm X\boldsymbol\beta + \mathrm G\mathbf u +
                \boldsymbol\epsilon,

where :math:`\mathbf u \sim \mathcal N(\mathbf 0, \sigma_g^2\mathrm I)` is a
vector of random effects and :math:`\epsilon_i` are iid Normal random
variables with variance :math:`\sigma_e^2` each.
The second part connects the latent variable to the observed one:

.. math::

    y_i ~|~ z_i \sim \text{ExpFam}(\mu_i = g(z_i)),

where :math:`g(\cdot)` is a link function and :math:`\text{ExpFam}(\cdot)` is
an exponential-family distribution.
The marginal likelihood is thus given by

.. math::

    p(\mathbf y) = \int
      \prod_i \text{ExpFam}(y_i ~|~ \mu_i = g(z_i))
      \mathcal N(\mathbf z ~|~ \mathrm X\boldsymbol\beta,
                 \sigma_g^2\mathrm G\mathrm G^{\intercal}
                 + \sigma_e^2\mathrm I)
    \mathrm d\mathbf z

This module implements the Expectation Propagation algorithm for parameter
fitting via Maximum Likelihood: :class:`.ExpFamEP`.

Public interface
^^^^^^^^^^^^^^^^
"""

from .glmm import GLMM

__all__ = ['GLMM']



"""
**************
Mean functions
**************

- :class:`.LinearMean`
- :class:`.OffsetMean`
- :class:`.SumMean`

Linear mean
^^^^^^^^^^^

.. autoclass:: LinearMean
  :members:

Offset mean
^^^^^^^^^^^

.. autoclass:: OffsetMean
  :members:

Sum mean
^^^^^^^^

.. autoclass:: SumMean
  :members:
"""

from .offset import OffsetMean
from .linear import LinearMean
from .sum import SumMean

__all__ = ['OffsetMean', 'LinearMean', 'SumMean']



import pytest

from numpy.random import RandomState
from numpy.testing import assert_allclose
from numpy import ascontiguousarray, sqrt, ones
from numpy_sugar.linalg import economic_qs, economic_qs_linear

from glimix_core.example import linear_eye_cov
from glimix_core.glmm import GLMM
from glimix_core.random import bernoulli_sample

from optimix import check_grad

def test_glmm():
    random = RandomState(0)
    X = random.randn(100, 5)
    K = linear_eye_cov().feed().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, 100)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMM((nsuc, ntri), 'binomial', X, QS)

    assert_allclose(glmm.value(), -272.1213895386019)
    assert_allclose(check_grad(glmm), 0, atol=1e-4)

def test_glmm_wrong_qs():
    random = RandomState(0)
    X = random.randn(10, 15)
    K = linear_eye_cov().feed().value()
    QS = [0, 1]

    ntri = random.randint(1, 30, 10)
    nsuc = [random.randint(0, i) for i in ntri]

    with pytest.raises(ValueError):
        print(GLMM((nsuc, ntri), 'binomial', X, QS))

def test_glmm_optimize():
    random = RandomState(0)
    X = random.randn(100, 5)
    K = linear_eye_cov().feed().value()
    QS = economic_qs(K)

    ntri = random.randint(1, 30, 100)
    nsuc = [random.randint(0, i) for i in ntri]

    glmm = GLMM((nsuc, ntri), 'binomial', X, QS)

    assert_allclose(glmm.value(), -272.1213895386019)
    glmm.fix('beta')
    glmm.fix('scale')

    glmm.feed().maximize(progress=False)
    assert_allclose(glmm.value(), -271.3681536575776)

    glmm.unfix('beta')
    glmm.unfix('scale')

    glmm.feed().maximize(progress=False)
    assert_allclose(glmm.value(), -264.67173062461313)

def test_glmm_bernoulli_problematic():
    random = RandomState(1)
    N = 500
    G = random.randn(N, N+50)
    y = bernoulli_sample(0.0, G, random_state=random)
    y = (y, )

    G = ascontiguousarray(G, dtype=float)
    _stdnorm(G, 0, out=G)
    G /= sqrt(G.shape[1])

    QS = economic_qs_linear(G)
    S0 = QS[1]
    S0 /= S0.mean()

    X = ones((len(y[0]), 1))
    model = GLMM(y, 'bernoulli', X, QS=(QS[0], QS[1]))
    model.delta = 0
    model.fix('delta')
    model.feed().maximize(progress=False)
    assert_allclose(model.feed().value(), -344.86474884323525)
    assert_allclose(model.delta, 0, atol=1e-6)
    assert_allclose(model.scale, 0.6025069154820977)
    assert_allclose(model.scale, 0.6025069154820977)
    assert_allclose(model.beta, [-0.018060946539989742])

def _stdnorm(X, axis=None, out=None):
    X = ascontiguousarray(X)
    if out is None:
        out = X.copy()

    m = out.mean(axis)
    s = out.std(axis)
    ok = s > 0

    out -= m

    if out.ndim == 1:
        if s > 0:
            out /= s
    else:
        out[..., ok] /= s[ok]

    return out



from __future__ import division

from numpy import log, pi
from numpy.linalg import slogdet, solve
from numpy_sugar import is_all_finite

from optimix import FunctionReduce


class GP(FunctionReduce):
    r"""Gaussian Process inference via maximum likelihood.

    Args:
        y (array_like): outcome variable.
        mean (:class:`optimix.Function`): mean function.
                                          (Refer to :doc:`mean`.)
        cov (:class:`optimix.Function`): covariance function.
                                         (Refer to :doc:`cov`.)

    Example
    -------

    .. doctest::

        >>> from numpy.random import RandomState
        >>>
        >>> from glimix_core.example import offset_mean
        >>> from glimix_core.example import linear_eye_cov
        >>> from glimix_core.gp import GP
        >>> from glimix_core.random import GPSampler
        >>>
        >>> random = RandomState(94584)
        >>>
        >>> mean = offset_mean()
        >>> cov = linear_eye_cov()
        >>>
        >>> y = GPSampler(mean, cov).sample(random)
        >>>
        >>> gp = GP(y, mean, cov)
        >>> print('Before: %.4f' % gp.feed().value())
        Before: -164.0064
        >>> gp.feed().maximize(progress=False)
        >>> print('After: %.4f' % gp.feed().value())
        After: -163.4192
    """

    def __init__(self, y, mean, cov):
        super(GP, self).__init__([mean, cov], name='GP')

        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the phenotype.")

        self._y = y
        self._cov = cov
        self._mean = mean

    def _lml_gradient_mean(self, mean, cov, gmean):
        Kiym = solve(cov, self._y - mean)
        return gmean.T.dot(Kiym)

    def _lml_gradient_cov(self, mean, cov, gcov):
        Kiym = solve(cov, self._y - mean)
        return (
            -solve(cov, gcov).diagonal().sum() + Kiym.dot(gcov.dot(Kiym))) / 2

    def value_reduce(self, values):  # pylint: disable=R0201
        mean = values['GP[0]']
        cov = values['GP[1]']
        ym = self._y - mean
        Kiym = solve(cov, ym)

        (s, logdet) = slogdet(cov)
        assert s == 1.

        n = len(self._y)
        return -(logdet + ym.dot(Kiym) + n * log(2 * pi)) / 2

    def gradient_reduce(self, values, gradients):
        mean = values['GP[0]']
        cov = values['GP[1]']
        gmean = gradients['GP[0]']
        gcov = gradients['GP[1]']

        grad = dict()
        for n, g in iter(gmean.items()):
            grad['GP[0].' + n] = self._lml_gradient_mean(mean, cov, g)

        for n, g in iter(gcov.items()):
            grad['GP[1].' + n] = self._lml_gradient_cov(mean, cov, g)

        return grad



from numpy import arange
from numpy.testing import assert_allclose

from glimix_core.mean import OffsetMean, LinearMean, SumMean

from optimix import check_grad


def test_summean_optimix():
    X = [[5.1, 1.0],
         [2.1, -0.2]]

    mean0 = LinearMean(2)
    mean0.set_data((X, ))
    mean0.effsizes = [-1.0, 0.5]

    mean1 = OffsetMean()
    mean1.set_data((arange(2), ))
    mean1.offset = 2.0

    mean = SumMean([mean0, mean1])

    assert_allclose(mean.feed().value(), [-2.6, -0.2])
    assert_allclose(check_grad(mean.feed()), 0, atol=1e-6)



r"""
****************
Gaussian Process
****************

Introduction
^^^^^^^^^^^^

A GP is a statistical distribution :math:`Y_t`, :math:`t\in\mathrm T`, for
which any finite linear combination of samples has a joint Gaussian
distribution `[1]`_ `[2]`_.
An instance of such class of processes is defined via a mean function
:math:`m(\cdot)` and a covariance function :math:`k(\cdot, \cdot)` whose
domains are :math:`\mathrm T` and :math:`\mathrm T\times\mathrm T`,
respectively.

An instance of such process is given by

.. math::

    \mathbf y \sim \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K)

Class :class:`.GP` performs inference over the mean and covariance
parameters via maximum likelihood and Expectation Propagation `[3]`_
approximation.

.. _[1]: https://en.wikipedia.org/wiki/Gaussian_process
.. _[2]: http://www.gaussianprocess.org/gpml/
.. _[3]: http://www.gaussianprocess.org/gpml/chapters/RW3.pdf


Public interface
^^^^^^^^^^^^^^^^
"""
from .gp import GP

__all__ = ['GP']



from __future__ import absolute_import, division, unicode_literals

import logging
from math import fsum

from numpy import dot, empty, inf, isfinite, log, maximum, zeros
from numpy.linalg import norm
from numpy_sugar import epsilon
from numpy_sugar.linalg import cho_solve, ddot, dotd

from .posterior import Posterior
from .site import Site

MAX_ITERS = 100
RTOL = epsilon.small * 1000
ATOL = epsilon.small * 1000


class EP(object):  # pylint: disable=R0903
    r"""Expectation Propagation algorithm.

    Let

    .. math::

        \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K)

    be the prior distribution.
    This class estimates the log of the marginal likelihood

    .. math::

        p(\mathbf y) = \int \prod_i p(y_i | \mu_i = g(z_i))
            \mathcal N(\mathbf z ~|~ \mathbf m, \mathrm K) \mathrm d\mathbf z

    via Expectation Propagation and provides its gradient.

    Attributes:
        _site (:class:`glimix_core.ep.site.Site`): site-likelihood.
        _psite (:class:`glimix_core.ep.site.Site`): previous
                                                        site-likelihood.
        _cav (:class:`glimix_core.ep.cavity.Cavity`): cavity distribution.
        _posterior (:class:`glimix_core.ep.posterior.Posterior`):
                                                        posterior distribution.
        _moments (dict): moments for KL moment matching.
    """

    def __init__(self):
        self._logger = logging.getLogger(__name__)

        self._site = None
        self._psite = None

        self._cav = None
        self._posterior = None

        self._moments = {'log_zeroth': None, 'mean': None, 'variance': None}

        self._need_params_update = True

    def _compute_moments(self):
        raise NotImplementedError

    def _initialize(self, mean, QS):
        self._logger.debug("EP parameters initialization.")
        self._need_params_update = True

        nsamples = len(mean)

        self._site = Site(nsamples)
        self._psite = Site(nsamples)

        self._cav = dict(tau=zeros(nsamples), eta=zeros(nsamples))
        self._posterior = Posterior(self._site)
        self._posterior.set_prior_mean(mean)
        self._posterior.set_prior_cov(QS)

        self._moments = {
            'log_zeroth': empty(nsamples),
            'mean': empty(nsamples),
            'variance': empty(nsamples)
        }

        self._posterior.initialize()

    def _lml(self):
        L = self._posterior.L()
        Q, S = self._posterior.prior_cov()
        ttau = self._site.tau
        teta = self._site.eta
        ctau = self._cav['tau']
        ceta = self._cav['eta']
        m = self._posterior.prior_mean()

        TS = ttau + ctau

        lml = [
            -log(L.diagonal()).sum(),
            -0.5 * sum(log(S)),
            # lml += 0.5 * sum(log(ttau)),
            +0.5 * dot(teta, dot(Q, cho_solve(L, dot(Q.T, teta)))),
            -0.5 * dot(teta, teta / TS),
            +dot(m, teta) - 0.5 * dot(m, ttau * m),
            -0.5 *
            dot(m * ttau, dot(Q, cho_solve(L, dot(Q.T, 2 * teta - ttau * m)))),
            +sum(self._moments['log_zeroth']),
            +0.5 * sum(log(TS)),
            # lml -= 0.5 * sum(log(ttau)),
            -0.5 * sum(log(ctau)),
            +0.5 * dot(ceta / TS, ttau * ceta / ctau - 2 * teta)
        ]
        lml = fsum(lml)

        if not isfinite(lml):
            raise ValueError("LML should not be %f." % lml)

        return lml

    def _lml_derivative_over_mean(self, dm):
        L = self._posterior.L()
        Q, _ = self._posterior.prior_cov()
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.prior_mean()

        dlml = dot(diff, dm)
        dlml -= dot(diff, dot(Q, cho_solve(L, dot(Q.T, (ttau * dm.T).T))))

        return dlml

    def _lml_derivative_over_cov(self, dQS):
        L = self._posterior.L()
        Q, _ = self._posterior.prior_cov()
        ttau = self._site.tau
        teta = self._site.eta

        diff = teta - ttau * self._posterior.prior_mean()

        v0 = dot(dQS[0], dQS[1] * dot(dQS[0].T, diff))
        v1 = ttau * dot(Q, cho_solve(L, dot(Q.T, diff)))
        dlml = 0.5 * dot(diff, v0)
        dlml -= dot(v0, v1)
        dlml += 0.5 * dot(v1, dot(dQS[0], dQS[1] * dot(dQS[0].T, v1)))
        dqs = ddot(dQS[1], dQS[0].T, left=True)
        diag = dotd(dQS[0], dqs)
        dlml -= 0.5 * sum(ttau * diag)

        tmp = cho_solve(L, dot(ddot(Q.T, ttau, left=False), dQS[0]))
        dlml += 0.5 * sum(ttau * dotd(Q, dot(tmp, dqs)))

        return dlml

    def _params_update(self):
        if not self._need_params_update:
            return

        self._logger.debug('EP parameters update loop has started.')

        i = 0
        tol = -inf
        while i < MAX_ITERS and norm(self._site.tau - self._psite.tau) > tol:
            self._psite.tau[:] = self._site.tau
            self._psite.eta[:] = self._site.eta

            self._cav['tau'][:] = maximum(self._posterior.tau - self._site.tau,
                                          0)
            self._cav['eta'][:] = self._posterior.eta - self._site.eta
            self._compute_moments()

            self._site.update(self._moments['mean'], self._moments['variance'],
                              self._cav)

            self._posterior.update()

            tol = RTOL * norm(self._psite.tau) + ATOL
            i += 1

        if i == MAX_ITERS:
            msg = ('Maximum number of EP iterations has' + ' been attained.')
            msg += " Last EP step was: %.10f." % norm(self._site.tau -
                                                      self._psite.tau)
            raise ValueError(msg)

        self._need_params_update = False
        self._logger.debug('EP loop has performed %d iterations.', i)
